import { Auth } from "@auth/core";
import { headers } from "next/headers";
import { NextResponse } from "next/server";
import { detectOrigin, reqWithEnvUrl } from "./env.js";
/** Server-side method to read the session. */
async function getSession(headers, config) {
    const origin = detectOrigin(headers);
    const request = new Request(`${origin}/api/auth/session`, {
        headers: { cookie: headers.get("cookie") ?? "" },
    });
    config.useSecureCookies ?? (config.useSecureCookies = origin.protocol === "https:");
    return Auth(request, {
        ...config,
        callbacks: {
            ...config.callbacks,
            // Since we are server-side, we don't need to filter out the session data
            // See https://nextjs.authjs.dev/v5#authenticating-server-side
            session({ session, user, token }) {
                return { ...session, user: user ?? token };
            },
        },
    });
}
function isReqWrapper(arg) {
    return typeof arg === "function";
}
export function initAuth(config) {
    return (...args) => {
        if (!args.length)
            return getSession(headers(), config).then((r) => r.json());
        if (args[0] instanceof Request) {
            // middleare.ts
            // export { auth as default } from "auth"
            const req = args[0];
            const ev = args[1];
            return handleAuth([req, ev], config);
        }
        if (isReqWrapper(args[0])) {
            // middleware.ts/router.ts
            // import { auth } from "auth"
            // export default auth((req) => { console.log(req.auth) }})
            const userMiddlewareOrRoute = args[0];
            return async (...args) => {
                return handleAuth(args, config, userMiddlewareOrRoute);
            };
        }
        // API Routes, getServerSideProps
        const request = "req" in args[0] ? args[0].req : args[0];
        const response = "res" in args[0] ? args[0].res : args[1];
        return getSession(
        // @ts-expect-error
        new Headers(request.headers), config).then(async (authResponse) => {
            const auth = await authResponse.json();
            // Preserve cookies set by Auth.js Core
            cloneSetCookie({ from: authResponse, to: response });
            return auth;
        });
    };
}
async function handleAuth(args, config, userMiddlewareOrRoute) {
    const request = reqWithEnvUrl(args[0]);
    const sessionResponse = await getSession(request.headers, config);
    const auth = await sessionResponse.json();
    // If we are handling a recognized NextAuth.js request,
    // don't require authorization to avoid an accidental redirect loop
    let authorized = true;
    // Infer basePath from NEXTAUTH_URL if provided, default to /api/auth
    const { pathname: basePath } = new URL(process.env.NEXTAUTH_URL ?? "http://a/api/auth");
    if (!isNextAuthAction(request, config, basePath) &&
        config.callbacks?.authorized) {
        authorized = await config.callbacks.authorized({ request, auth });
    }
    let response = NextResponse.next?.();
    if (authorized instanceof Response) {
        // User returned a custom response, like redirecting to a page or 401, respect it
        response = authorized;
    }
    else if (userMiddlewareOrRoute) {
        // Execute user's middleware/handler with the augmented request
        const augmentedReq = request;
        augmentedReq.auth = auth;
        response =
            // @ts-expect-error
            (await userMiddlewareOrRoute(augmentedReq, args[1])) ??
                NextResponse.next();
    }
    else if (!authorized) {
        const signInPage = config.pages?.signIn ?? "/api/auth/signin";
        if (request.nextUrl.pathname !== signInPage) {
            // Redirect to signin page by default if not authorized
            request.nextUrl.pathname = signInPage;
            request.nextUrl.searchParams.set("callbackUrl", request.nextUrl.href);
            response = NextResponse.redirect(request.nextUrl);
        }
    }
    const finalResponse = new Response(response?.body, response);
    // Preserve cookies set by Auth.js Core
    cloneSetCookie({ from: sessionResponse, to: finalResponse });
    return finalResponse;
}
/**
 * Mutates `to` with the set-cookie from `from`
 * Uses `getSetCookie` if available, otherwise falls back to `set-cookie`
 *
 * getSetCookie is missing from the types, but it's available on many platforms.
 */
function cloneSetCookie({ from, to }) {
    const authCookies = "getSetCookie" in from.headers
        ? from.headers.getSetCookie()
        : from.headers.get("set-cookie");
    if (!authCookies)
        return;
    if (Array.isArray(authCookies)) {
        authCookies.forEach((cookie) => to.headers.append("set-cookie", cookie));
    }
    else {
        to.headers.set("set-cookie", authCookies);
    }
}
/** Check if the request is for a NextAuth.js action. */
function isNextAuthAction(req, config, basePath) {
    const { pathname } = req.nextUrl;
    const action = pathname.replace(`${basePath}/`, "");
    const pages = Object.values(config.pages ?? {});
    return actions.has(action) || pages.some((page) => pathname === page);
}
const actions = new Set([
    "providers",
    "session",
    "csrf",
    "signin",
    "signout",
    "callback",
    "verify-request",
    "error",
]);
